/*
INTCHAIN.C 
Andrew Schulman, May 1993
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)
Copyright (C) 1993 Andrew Schulman.  All rights reserved.

bcc intchain.c map.c

Uses single-step to trace through interrupt chains
usage:   intchain intno/ax/bx/cx/dx
example: intchain 21/6200

Modified 8/20/94:
-- Display ARPL instruction (Windows V86 breakpoint)
-- Print out any INT XXh AX=XXXXh encountered

To do:
In addition to logging when (seg != prev_seg), also check if address
is in map.  Would pick up e.g. device-driver Strat and Intr.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

#ifdef __cplusplus
typedef void interrupt (far *INTRFUNC)(...);
#else
typedef void (interrupt far *INTRFUNC)(void);
#endif

typedef void (far *FARFUNC)(void);

#ifndef MK_FP
#define MK_FP(s,o)  ((((DWORD) s) << 16) + (o))
#endif

#define MK_LIN(fp)  ((((DWORD) FP_SEG(fp)) << 4) + FP_OFF(fp))

#pragma pack(1)

typedef struct {
#ifdef __TURBOC__
    WORD bp,di,si,ds,es,dx,cx,bx,ax;
#else
    WORD es,ds,di,si,bp,sp,bx,dx,cx,ax;     /* same as PUSHA */
#endif
    WORD ip,cs,flags;
    } REG_PARAMS;
    
#define INT_INSTR       0xCD
#define ARPL			0x63
#define BREAKPOINT		0xCC
#define TRACE_FLAG      0x100

extern char *find_owner(DWORD lin_addr);    // in map.c

void fail(const char *s) { puts(s); exit(1); }

#define MAX_ADDR        512

static BYTE prev_inst = 0;
static WORD weird = 0;
static WORD intax[MAX_ADDR] = {0};	// AX at time of INT
static WORD volatile instr = 0, int_instr = 0;
static WORD prev_seg = 0, my_seg = 0;
static void far * *addr;
static int num_addr = 0;

INTRFUNC my_getvect(unsigned intno)
{
    INTRFUNC far *ivt = (INTRFUNC far *) 0L;
    return ivt[intno];
}

void interrupt far single_step(REG_PARAMS r)    // INT 1 handler
{
    WORD seg;
    BYTE far *fp;
    
    if ((seg = r.cs) == my_seg)
        return;
    
    fp = (BYTE far *) MK_FP(r.cs, r.ip);
    if (fp[0] == INT_INSTR)
	{
#ifdef BRKPT_INT		
		BYTE far *target = (BYTE far *) my_getvect(fp[1]);
		prev_inst = *target;
		*target = BREAKPOINT;
#endif		
		intax[int_instr++] = r.ax;
	}
    instr++;
        
    if ((seg != prev_seg) || (fp[0] == ARPL) || (fp[0] == INT_INSTR))
    {
        if (num_addr < MAX_ADDR)
            addr[num_addr++] = (void far *) fp;
        prev_seg = seg;
    }
}

#define GET_FLAGS(reg)  _asm { pushf } ; _asm { pop reg }
#define SET_FLAGS(reg)  _asm { push reg } ; _asm { popf }

void set_flag(unsigned mask)
{
    GET_FLAGS(ax);
    _asm or ax, word ptr mask
    SET_FLAGS(ax);
}

void clear_flag(unsigned mask)
{
    GET_FLAGS(ax);
    _asm mov bx, word ptr mask
    _asm not bx
    _asm and ax, bx
    SET_FLAGS(ax);
}

#ifdef BRKPT_INT
void interrupt far breakpoint(REG_PARAMS r)    // INT 3 handler
{
    WORD seg;
    BYTE far *fp = (BYTE far *) MK_FP(r.cs, r.ip);
    if (fp[0] != BREAKPOINT)
		weird++;
	fp[0] = prev_inst;
	set_flag(TRACE_FLAG);
}
#endif

FARFUNC get_xms(void)
{
    _asm mov ax, 4300h
    _asm int 2fh
    _asm cmp al, 80h
    _asm je present
absent:
    fail("XMS not present!");
present:
    _asm mov ax, 4310h
    _asm int 2fh
    _asm mov ax, bx
    _asm mov dx, es
    // retval in DX:AX
}

main(int argc, char *argv[])
{
    static int intrfunc = 0;    /* make sure not in a register */
    INTRFUNC old_sstep, old_brkpt;
    FARFUNC func = (FARFUNC) 0;
    FARFUNC xms_func = (FARFUNC) 0;
    void far *fp;
    char *s;
    WORD intno, _ax, _bx, _cx, _dx;
    int a20off = 0;
    int i, j;

    puts("INTCHAIN 1.10 -- Walks interrupt chains");
    puts("From \"Undocumented DOS\", 2nd edition (Addison-Wesley, 1993)");
    puts("Copyright (C) 1993-94 Andrew Schulman.  All rights reserved.\n");
    
    if (argc < 2)
        fail("usage: intchain [-a20off] <intno|xms|seg:ofs>/ax/bx/cx/dx");
    
    if (strcmp(strupr(argv[1]), "-A20OFF") == 0)
    {
        xms_func = get_xms();
        a20off++;
        argv++;
    }
    
    if (strncmp(strupr(argv[1]), "XMS", 3) == 0)
    {
        func = get_xms();
        sscanf(argv[1], "XMS/%04X/%04X/%04X/%04X",
            &_ax, &_bx, &_cx, &_dx);
        printf("Tracing XMS at %Fp\n", func);
    }
    else if (strchr(argv[1], ':'))
    {
        WORD seg, ofs;
        sscanf(argv[1], "%04X:%04X/%04X/%04X/%04X/%04X",
            &seg, &ofs, &_bx, &_cx, &_dx);
        func = (FARFUNC) MK_FP(seg, ofs);
        printf("Tracing function at %Fp\n", func);
    }
    else
    {
        sscanf(argv[1], "%02X/%04X/%04X/%04X/%04X",
            &intno, &_ax, &_bx, &_cx, &_dx);

        /* single-step doesn't go through INT, so turn the INT into
           a PUSHF and far CALL */
        if (! (func = (FARFUNC) _dos_getvect(intno)))
            fail("INT unused");
        intrfunc++;     // so do PUSHF when call func
        printf("Tracing INT %02X AX=%04X\n", intno, _ax);
    }
    
    if (! (addr = (void far **) calloc(MAX_ADDR, sizeof(void far *))))
        fail("insufficient memory");
    fp = (void far *) main;
    my_seg = FP_SEG(fp);
    
    old_sstep = _dos_getvect(1);
    _dos_setvect(1, (INTRFUNC) single_step);
#ifdef BRKPT_INT	
	old_brkpt = _dos_getvect(3);
    _dos_setvect(3, (INTRFUNC) breakpoint);
#endif	
    
    if (a20off)
    {
        _asm mov ah, 6
        (*xms_func)();  // local disable A20 line
    }
    
    set_flag(TRACE_FLAG);

    /* call the code */
    _asm mov ax, _ax
    _asm mov bx, _bx
    _asm mov cx, _cx
    _asm mov dx, _dx
    if (intrfunc)
        _asm pushf
    (*func)();
            
    clear_flag(TRACE_FLAG);
    _dos_setvect(1, old_sstep);
#ifdef BRKPT_INT	
	_dos_setvect(3, old_brkpt);
#endif
    
    printf("%u instructions\n", instr);
    if (int_instr)
#ifdef BRKPT_INT		
        printf("Encountered %u INT\n", int_instr);
#else	
        printf("Skipped %u INT\n", int_instr);
#endif	
    printf("\n");
    
    for (i=0, j=0; i<num_addr; i++)
    {
        s = find_owner(MK_LIN(addr[i]));
        printf("%Fp\t%s", addr[i], s? s: " ");
		if (*((BYTE far *) addr[i]) == ARPL)
			printf("\tARPL");
		else if (*((BYTE far *) addr[i]) == INT_INSTR)
		{
			printf("\tINT %02Xh AX=%04Xh", 
				*((BYTE far *) addr[i]+1),
				intax[j]);
			j++; 
			if (j > int_instr)
				fail("Something wrong with INT tracking!");
		}
		printf("\n");
    }
    if (num_addr == MAX_ADDR)
        fail("Overflow: very long INT chain!");
    
    return 0;
}
